#ifndef LOBBY_H
#define LOBBY_H

#include "csdl.h"
#include "boxerhandler.h"
#include "snet.h"

#define LOBBY_CODE 0x1234abcd

#define MAX_PLAYER 64
#define MAX_ROOM 8
#define MAX_PACKET_SIZE 256

#define	LOBBY_CONNECT   0x01
#define	SEND_CODE		0x02
#define	CODE_SUCCESS	0x03
#define	REQUEST_STATES	0x04
#define	SEND_STATES		0x05
#define	REQUEST_ROOM	0x06
#define MATCH_RED		0x07
#define MATCH_BLUE		0x08
#define RETURN_LOBBY	0x09
#define	LEAVE_LOBBY		0x0a
#define	LOBBY_FULL		0xff
#define	LOBBY_TIMEOUT	0xfe
#define	CODE_ERROR		0xfd
#define	KILL_CLIENT		0xfc
#define	SOCKET2_ERROR	0xfb

#define NO_LOBBY		0x00
#define LOBBY_CERTIFY	0x01
#define LOBBY_WAIT		0x02
#define LOBBY_ACCEPT	0x03
#define LOBBY_MATCH		0x04

#define EMPTY_ROOM		0xff
#define MATCH_ROOM		0xfe


class IP;
class SocketSet;

class TCP_Socket {
public :
	TCP_Socket();
	~TCP_Socket();
	int Open(IP &ip);
	void Close();
	void Destroy() {socket = NULL;}
	bool WasOpen();
	
	void Accept(TCP_Socket &new_socket) {
		new_socket.socket = SDLNet_TCP_Accept(socket);	
	}		

	int IsReady() {return SDLNet_SocketReady(socket);}
	int GetOnce(void *data, int maxlen);
	int SendData(void *data, int maxlen);
	int GetData(void *data, int maxlen);
	void Send8(Uint8 data);
	Uint8 Get8();
	void Send16(Uint16 data);
	Uint16 Get16();
	void Send32(Uint32 data);
	Uint32 Get32();
private:
	TCPsocket socket;
	friend class IP;
	friend class SocketSet;	
};

class UDP_Socket {
public :
	UDP_Socket();
	~UDP_Socket();
	int Open(Uint16 port);
	void Close();
	bool WasOpen();

	void SetDestination(IP &ip);
	void SetDestination(Uint32 host, Uint16 port);
    void Unbind() {SDLNet_UDP_Unbind(socket, 1);}
	
	int ResizePacket(int size);

	int IsReady() {return SDLNet_SocketReady(socket);}
	int SendData(void *data, int maxlen);
	int GetData(void *data, int maxlen);
	void Send8(Uint8 data);
	Uint8 Get8();
	void Send16(Uint16 data);
	Uint16 Get16();
	void Send32(Uint32 data);
	Uint32 Get32();	
private:
	UDPsocket socket;
	UDPpacket *packet;
	friend class IP;
	friend class SocketSet;	
};

class IP {
public:
	int ResolveHost(const char *host, Uint16 port) {
		return SDLNet_ResolveHost(&ip, host, port);	
	}
	const char *HostName() {
		return SDLNet_ResolveIP(&ip);
	}
	TCPsocket TCP_Open() {
		return SDLNet_TCP_Open(&ip);	
	}
	void TCP_GetAddress(TCP_Socket &tcp) {
		ip = *SDLNet_TCP_GetPeerAddress(tcp.socket);
	}
	void UDP_GetAddress(UDP_Socket &udp) {
		ip = *SDLNet_UDP_GetPeerAddress(udp.socket, -1);	
	}    

	IPaddress Address() const {return ip;}
	void SetHost(Uint32 host) {ip.host = host;}
	void SetPort(Uint16 port) {ip.port = port;}
	Uint32 Host() const {return ip.host;}
	Uint16 Port() const {return ip.port;}

	Uint16 PortR16() const {return SDLNet_Read16(&ip.port);}
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	int IP1() {return 0xff & (ip.host >> 24);}
	int IP2() {return 0xff & (ip.host >> 16);}
	int IP3() {return 0xff & (ip.host >> 8);}
	int IP4() {return 0xff & (ip.host);}
#else
	int IP1() {return 0xff & (ip.host);}
	int IP2() {return 0xff & (ip.host >> 8);}
	int IP3() {return 0xff & (ip.host >> 16);}
	int IP4() {return 0xff & (ip.host >> 24);}
#endif
private:
	IPaddress ip;
};

class SocketSet {
public:
	SocketSet();
	~SocketSet();
	int Create(int number);
	void Free();
	bool WasCreate();

	int Number();
	int Max();

	int Check(Uint32 timeout = 0) {
		return SDLNet_CheckSockets(set, timeout);	
	}

	void Add_TCP_Socket(TCP_Socket &tcp);
	void Del_TCP_Socket(TCP_Socket &tcp);

	void Add_UDP_Socket(UDP_Socket &udp);
	void Del_UDP_Socket(UDP_Socket &udp);

private:
	SDLNet_SocketSet set;
};


class Lobby {
public :
	Lobby();
	~Lobby();
	int Init();

	void Quit();
	bool WasInit();

	Uint8 PlayerNumber() const {return player_number;}
	Uint8 ClientState(int i) {return client_state[i];}
	Uint8 *ClientStates() {return client_state;}
	Uint8 RoomState(int i) {return room_state[i];}
	Uint8* RoomName(int i) {return client_name[room_state[i]];}
	Uint8 *RoomStates() {return room_state;}
	
	int ResolveHost(Uint16 port);

	void CheckSockets();
	void WaitClients();
	void HandleClients();
	int AddClient();
	void HandleClient(int number);
	void SendState(int number);

private :
	SocketSet socket_set;
	TCP_Socket server_socket;
	TCP_Socket client_socket[MAX_PLAYER];
	IP server_ip;
	IP client_ip[MAX_PLAYER];

	Uint8 client_name[MAX_PLAYER][NAME_LENGTH+1];
	Uint8 client_state[MAX_PLAYER];
	Uint16 client_port[MAX_PLAYER];
	Uint32 wait_count[MAX_PLAYER];

	Uint8 room_state[MAX_ROOM];
	Uint8 player_number;

	FILE* lobby_log;
};


class LobbyClient {
public :
	LobbyClient();
	~LobbyClient();
	int Init();

	void Quit();
	bool WasInit();

	int ConnectLobby(char* user_name, char* hostname, Uint16 port, Uint16 port_tcp);	
	void SendLobbyCode() {
		tcp_socket.Send8(SEND_CODE);		
		tcp_socket.Send32(LOBBY_CODE);
	}
	void LeaveLobby() {
		tcp_socket.Send8(LEAVE_LOBBY);	
	}	
	void RequestState() {
		tcp_socket.Send8(REQUEST_STATES);
	}
	void RequestRoom(Uint8 number) {
		tcp_socket.Send8(REQUEST_ROOM);
		tcp_socket.Send8(number);
	}
	void ReturnLobby() {
		tcp_socket.Send8(RETURN_LOBBY);
	}
	
	void GetState();
	Uint8 PlayerNumber() const {return player_number;}
	Uint8 RoomState(int i) {return room_state[i];}
	Uint8 *RoomName(int i) {return room_name[i];}
	Uint8 *RoomStates() {return room_state;}

	int IsReady() {
		socket_set.Check(0);
		return tcp_socket.IsReady();
	}
	int WaitData(int max_count);
	Uint8 Handle();
		
	Uint16 Port() {return port;}
	char* HostIP() {
		sprintf(hostip, "%d.%d.%d.%d", ip.IP1(), ip.IP2(), ip.IP3(), ip.IP4()); 
		return hostip;	
	} 

private :
	SocketSet socket_set;
	TCP_Socket tcp_socket;
	IP client_ip, my_ip, ip;

	Uint8 room_state[MAX_ROOM];
	Uint8 room_name[MAX_ROOM][NAME_LENGTH+1];

	Uint8 player_number;
	Uint16 port;
	Uint32 host;
	char hostip[50];
};


class LobbyClientUDP {
public :
	LobbyClientUDP();
	~LobbyClientUDP();
	int Init();

	void Quit();
	bool WasInit();

	int ConnectLobby(char* user_name, char* hostname, Uint16 port, Uint16 port_udp);	
	void SendLobbyCode() {
		tcp_socket.Send8(SEND_CODE);		
		tcp_socket.Send32(LOBBY_CODE);
	}
	void LeaveLobby() {
		tcp_socket.Send8(LEAVE_LOBBY);	
	}	
	void RequestState() {
		tcp_socket.Send8(REQUEST_STATES);
	}
	void RequestRoom(Uint8 number) {
		tcp_socket.Send8(REQUEST_ROOM);
		tcp_socket.Send8(number);
	}
	void ReturnLobby() {
		udp_socket.Unbind();
		tcp_socket.Send8(RETURN_LOBBY);
	}
	
	void GetState();
	Uint8 PlayerNumber() const {return player_number;}
	Uint8 RoomState(int i) {return room_state[i];}
	Uint8 *RoomName(int i) {return room_name[i];}
	Uint8 *RoomStates() {return room_state;}

	int IsReady() {
		socket_set.Check(0);
		return tcp_socket.IsReady();
	}
	int IsUDPReady() {
		socket_set.Check(0);
		return udp_socket.IsReady();
	}
	int WaitData(int max_count);
	int WaitUDPData(int max_count);
	Uint8 Handle();
		
	void SetDestination(IP &ip) {
		udp_socket.SetDestination(ip);	
	}
	void SetDestination(Uint32 host, Uint16 port) {
		udp_socket.SetDestination(host, port);	
	}
	int SendLargeData(void *data, int maxlen);
	int GetLargeData(void *data, int maxlen);
	int SendData(void *data, int maxlen) {
		return udp_socket.SendData(data, maxlen);
	}
	int GetData(void *data, int maxlen) {
		return udp_socket.GetData(data, maxlen);
	}
	void Send8(Uint8 data) {udp_socket.Send8(data);}
	Uint8 Get8() {return udp_socket.Get8();}
	void Send16(Uint16 data) {udp_socket.Send16(data);}
	Uint16 Get16() {return udp_socket.Get16();}
	void Send32(Uint32 data) {udp_socket.Send32(data);}
	Uint32 Get32() {return udp_socket.Get32();}

	bool IsRedCorner() const {return (my_corner == MATCH_RED);}
	bool IsBlueCorner() const {return (my_corner == MATCH_BLUE);}
	void SetCornerRed() {my_corner = MATCH_RED;}
	void SetCornerBlue() {my_corner = MATCH_BLUE;}			
private :
	SocketSet socket_set;
	TCP_Socket tcp_socket;
	UDP_Socket udp_socket;
	IP client_ip, my_ip;

	Uint8 room_state[MAX_ROOM];
	Uint8 room_name[MAX_ROOM][NAME_LENGTH+1];

	Uint8 player_number;
	Uint8 my_corner;
};


class Lobby2 {
public :
	Lobby2();
	~Lobby2();
	int Init();

	void Quit();
	bool WasInit();

	Uint8 PlayerNumber() const {return player_number;}
	Uint8 ClientState(int i) {return client_state[i];}
	Uint8 *ClientStates() {return client_state;}
	Uint8 RoomState(int i) {return room_state[i];}
	Uint8* RoomName(int i) {return client_name[room_state[i]];}
	Uint8 *RoomStates() {return room_state;}
	
	int ResolveHost(Uint16 port);

	void CheckSockets();
	void WaitClients();
	void HandleClients();
	int AddClient();
	void HandleClient(int number);
	void SendState(int number);

private :
	SocketSet socket_set;
	TCP_Socket server_socket;
	TCP_Socket client_socket[MAX_PLAYER];
	TCP_Socket client_socket2[MAX_PLAYER];
	IP server_ip;
	IP client_ip[MAX_PLAYER];

	Uint8 player_number;
	Uint8 room_state[MAX_ROOM];
	Uint8 client_name[MAX_PLAYER][NAME_LENGTH+1];
	Uint8 client_state[MAX_PLAYER];
	Uint8 client_cont[MAX_PLAYER];
	Uint32 wait_count[MAX_PLAYER];

	FILE* lobby_log;
};

class LobbyClient2 {
public :
	LobbyClient2();
	~LobbyClient2();
	int Init();

	void Quit();
	bool WasInit();

	int ConnectLobby(char* user_name, char* hostname, Uint16 port);	
	void SendLobbyCode() {
		tcp_socket.Send8(SEND_CODE);		
		tcp_socket.Send32(LOBBY_CODE);
	}
	void LeaveLobby() {
		tcp_socket.Send8(LEAVE_LOBBY);	
	}	
	void RequestState() {
		tcp_socket.Send8(REQUEST_STATES);
	}
	void RequestRoom(Uint8 number) {
		tcp_socket.Send8(REQUEST_ROOM);
		tcp_socket.Send8(number);
	}
	void ReturnLobby() {
//		udp_socket.Unbind();
		tcp_socket.Send8(RETURN_LOBBY);
	}
	
	void GetState();
	Uint8 PlayerNumber() const {return player_number;}
	Uint8 RoomState(int i) {return room_state[i];}
	Uint8 *RoomName(int i) {return room_name[i];}
	Uint8 *RoomStates() {return room_state;}

	int IsReady() {
		socket_set.Check(0);
		return tcp_socket.IsReady();
	}
	int IsReady2() {
		socket_set.Check(0);
		return tcp_socket2.IsReady();
	}
/*	int IsUDPReady() {
		socket_set.Check(0);
		return udp_socket.IsReady();
	}  */
	int WaitData(int max_count);
	int WaitData2(int max_count);
//	int WaitUDPData(int max_count);
	Uint8 Handle();
		
/*	void SetDestination(IP &ip) {
		udp_socket.SetDestination(ip);	
	}  
	void SetDestination(Uint32 host, Uint16 port) {
		udp_socket.SetDestination(host, port);	
	}  */
//	int SendLargeData(void *data, int maxlen);
//	int GetLargeData(void *data, int maxlen);
	int SendData(void *data, int maxlen) {
		return tcp_socket.SendData(data, maxlen);
	}
	int GetData(void *data, int maxlen) {
		return tcp_socket.GetData(data, maxlen);
	}
	void Send8(Uint8 data) {tcp_socket.Send8(data);}
	Uint8 Get8() {return tcp_socket.Get8();}
	void Send16(Uint16 data) {tcp_socket.Send16(data);}
	Uint16 Get16() {return tcp_socket.Get16();}
	void Send32(Uint32 data) {tcp_socket.Send32(data);}
	Uint32 Get32() {return tcp_socket.Get32();}

	int SendPacket(void *data, int maxlen) {
		return tcp_socket2.SendData(data, maxlen);
	}
	int GetPacket(void *data, int maxlen) {
		return tcp_socket2.GetData(data, maxlen);
	}
	void SendInt(Uint32 data) {tcp_socket2.Send32(data);}
	Uint32 GetInt() {return tcp_socket2.Get32();}

	bool IsRedCorner() const {return (my_corner == MATCH_RED);}
	bool IsBlueCorner() const {return (my_corner == MATCH_BLUE);}
	void SetCornerRed() {my_corner = MATCH_RED;}
	void SetCornerBlue() {my_corner = MATCH_BLUE;}			
private :
	SocketSet socket_set;
	TCP_Socket tcp_socket;
	TCP_Socket tcp_socket2;
//	UDP_Socket udp_socket;
	IP server_ip;
	IP client_ip, my_ip;

	Uint8 player_number;
	Uint8 my_corner;

	Uint8 room_state[MAX_ROOM];
	Uint8 room_name[MAX_ROOM][NAME_LENGTH+1];
};

#endif // LOBBY_H
